@@ -20,6 +20,7 @@
 
 #include "image.h"
 #include "decctx.h"
+#include "encoder/encoder-context.h"
 
 #include <stdlib.h>
 #include <string.h>
@@ -38,6 +39,8 @@
 #define MEMORY_PADDING  0
 #endif
 
+#define STANDARD_ALIGNMENT 16
+
 #ifdef HAVE___MINGW_ALIGNED_MALLOC
 #define ALLOC_ALIGNED(alignment, size)         __mingw_aligned_malloc((size), (alignment))
 #define FREE_ALIGNED(mem)                      __mingw_aligned_free((mem))
@@ -62,22 +65,81 @@ static inline void *ALLOC_ALIGNED(size_t alignment, size_t size) {
 
 static const int alignment = 16;
 
+LIBDE265_API void* de265_alloc_image_plane(struct de265_image* img, int cIdx,
+                                           void* inputdata, int inputstride, void *userdata)
+{
+  int alignment = STANDARD_ALIGNMENT;
+  int stride = (img->get_width(cIdx) + alignment-1) / alignment * alignment;
+  int height = img->get_height(cIdx);
+
+  uint8_t* p = (uint8_t *)ALLOC_ALIGNED_16(stride * height + MEMORY_PADDING);
+
+  if (p==NULL) { return NULL; }
+
+  img->set_image_plane(cIdx, p, stride, userdata);
+
+  // copy input data if provided
+
+  if (inputdata != NULL) {
+    if (inputstride == stride) {
+      memcpy(p, inputdata, stride*height);
+    }
+    else {
+      for (int y=0;y<height;y++) {
+        memcpy(p+y*stride, ((char*)inputdata) + inputstride*y, inputstride);
+      }
+    }
+  }
+
+  return p;
+}
+
+
+LIBDE265_API void de265_free_image_plane(struct de265_image* img, int cIdx)
+{
+  uint8_t* p = (uint8_t*)img->get_image_plane(cIdx);
+  assert(p);
+  FREE_ALIGNED(p);
+}
+
 
 static int  de265_image_get_buffer(de265_decoder_context* ctx,
                                    de265_image_spec* spec, de265_image* img, void* userdata)
 {
-  int luma_stride   = (spec->width   + spec->alignment-1) / spec->alignment * spec->alignment;
-  int chroma_stride = (spec->width/2 + spec->alignment-1) / spec->alignment * spec->alignment;
+  const int rawChromaWidth  = spec->width  / img->sps.SubWidthC;
+  const int rawChromaHeight = spec->height / img->sps.SubHeightC;
+
+  int luma_stride   = (spec->width    + spec->alignment-1) / spec->alignment * spec->alignment;
+  int chroma_stride = (rawChromaWidth + spec->alignment-1) / spec->alignment * spec->alignment;
+
+  assert(img->sps.BitDepth_Y >= 8 && img->sps.BitDepth_Y <= 16);
+  assert(img->sps.BitDepth_C >= 8 && img->sps.BitDepth_C <= 16);
+
+  int luma_bpl   = luma_stride   * ((img->sps.BitDepth_Y+7)/8);
+  int chroma_bpl = chroma_stride * ((img->sps.BitDepth_C+7)/8);
 
   int luma_height   = spec->height;
-  int chroma_height = (spec->height+1)/2;
+  int chroma_height = rawChromaHeight;
+
+  bool alloc_failed = false;
 
   uint8_t* p[3] = { 0,0,0 };
-  p[0] = (uint8_t *)ALLOC_ALIGNED_16(luma_stride   * luma_height   + MEMORY_PADDING);
-  p[1] = (uint8_t *)ALLOC_ALIGNED_16(chroma_stride * chroma_height + MEMORY_PADDING);
-  p[2] = (uint8_t *)ALLOC_ALIGNED_16(chroma_stride * chroma_height + MEMORY_PADDING);
+  p[0] = (uint8_t *)ALLOC_ALIGNED_16(luma_height   * luma_bpl   + MEMORY_PADDING);
+  if (p[0]==NULL) { alloc_failed=true; }
 
-  if (p[0]==NULL || p[1]==NULL || p[2]==NULL) {
+  if (img->get_chroma_format() != de265_chroma_mono) {
+    p[1] = (uint8_t *)ALLOC_ALIGNED_16(chroma_height * chroma_bpl + MEMORY_PADDING);
+    p[2] = (uint8_t *)ALLOC_ALIGNED_16(chroma_height * chroma_bpl + MEMORY_PADDING);
+
+    if (p[1]==NULL || p[2]==NULL) { alloc_failed=true; }
+  }
+  else {
+    p[1] = NULL;
+    p[2] = NULL;
+    chroma_stride = 0;
+  }
+
+  if (alloc_failed) {
     for (int i=0;i<3;i++)
       if (p[i]) {
         FREE_ALIGNED(p[i]);
@@ -98,8 +160,9 @@ static void de265_image_release_buffer(de265_decoder_context* ctx,
 {
   for (int i=0;i<3;i++) {
     uint8_t* p = (uint8_t*)img->get_image_plane(i);
-    assert(p);
-    FREE_ALIGNED(p);
+    if (p) {
+      FREE_ALIGNED(p);
+    }
   }
 }
 
@@ -128,6 +191,9 @@ de265_image::de265_image()
   removed_at_picture_id = 0; // picture not used, so we can assume it has been removed
 
   decctx = NULL;
+  encctx = NULL;
+
+  encoder_image_release_func = NULL;
 
   //alloc_functions.get_buffer = NULL;
   //alloc_functions.release_buffer = NULL;
@@ -165,15 +231,25 @@ de265_image::de265_image()
 
 de265_error de265_image::alloc_image(int w,int h, enum de265_chroma c,
                                      const seq_parameter_set* sps, bool allocMetadata,
-                                     decoder_context* ctx, de265_PTS pts, void* user_data,
-                                     bool isOutputImage)
+                                     decoder_context* dctx,
+                                     encoder_context* ectx,
+                                     de265_PTS pts, void* user_data,
+                                     bool useCustomAllocFunc)
 {
-  if (allocMetadata) { assert(sps); }
+  //if (allocMetadata) { assert(sps); }
+  assert(sps);
+
+  this->sps = *sps;
+
+  release(); /* TODO: review code for efficient allocation when arrays are already
+                allocated to the requested size. Without the release, the old image-data
+                will not be freed. */
 
   ID = s_next_image_ID++;
   removed_at_picture_id = std::numeric_limits<int32_t>::max();
 
-  decctx = ctx;
+  decctx = dctx;
+  encctx = ectx;
 
   // --- allocate image buffer ---
 
@@ -209,17 +285,27 @@ de265_error de265_image::alloc_image(int w,int h, enum de265_chroma c,
 
   case de265_chroma_422:
     spec.format = de265_image_format_YUV422P8;
-    chroma_height = (chroma_height+1)/2;
+    chroma_width = (chroma_width+1)/2;
+    break;
+
+  case de265_chroma_444:
+    spec.format = de265_image_format_YUV444P8;
+    break;
+
+  case de265_chroma_mono:
+    spec.format = de265_image_format_mono8;
+    chroma_width = 0;
+    chroma_height= 0;
     break;
 
   default:
-    assert(false); // TODO: not implemented yet
+    assert(false);
     break;
   }
 
   spec.width  = w;
   spec.height = h;
-  spec.alignment = 16;
+  spec.alignment = STANDARD_ALIGNMENT;
 
 
   // conformance window cropping
@@ -243,29 +329,55 @@ de265_error de265_image::alloc_image(int w,int h, enum de265_chroma c,
   spec.visible_height= height_confwin;
 
 
+  bpp_shift[0] = (sps->BitDepth_Y > 8) ? 1 : 0;
+  bpp_shift[1] = (sps->BitDepth_C > 8) ? 1 : 0;
+  bpp_shift[2] = bpp_shift[1];
+
+
   // allocate memory and set conformance window pointers
 
-  void* alloc_userdata = decctx->param_image_allocation_userdata;
-  if (isOutputImage) {
+  void* alloc_userdata = NULL;
+  if (decctx) alloc_userdata = decctx->param_image_allocation_userdata;
+  if (encctx) alloc_userdata = encctx->param_image_allocation_userdata; // actually not needed
+
+  if (encctx && useCustomAllocFunc) {
+    encoder_image_release_func = encctx->release_func;
+
+    // if we do not provide a release function, use our own
+
+    if (encoder_image_release_func == NULL) {
+      image_allocation_functions = de265_image::default_image_allocation;
+    }
+    else {
+      image_allocation_functions.get_buffer     = NULL;
+      image_allocation_functions.release_buffer = NULL;
+    }
+  }
+  else if (decctx && useCustomAllocFunc) {
     image_allocation_functions = decctx->param_image_allocation_functions;
   }
   else {
     image_allocation_functions = de265_image::default_image_allocation;
   }
-  bool mem_alloc_success = image_allocation_functions.get_buffer(decctx, &spec, this,
-                                                                 alloc_userdata);
 
-  pixels_confwin[0] = pixels[0] + left*WinUnitX + top*WinUnitY*stride;
-  pixels_confwin[1] = pixels[1] + left + top*chroma_stride;
-  pixels_confwin[2] = pixels[2] + left + top*chroma_stride;
+  bool mem_alloc_success = true;
+
+  if (image_allocation_functions.get_buffer != NULL) {
+    mem_alloc_success = image_allocation_functions.get_buffer(decctx, &spec, this,
+                                                              alloc_userdata);
+
+    pixels_confwin[0] = pixels[0] + left*WinUnitX + top*WinUnitY*stride;
+    pixels_confwin[1] = pixels[1] + left + top*chroma_stride;
+    pixels_confwin[2] = pixels[2] + left + top*chroma_stride;
 
 
-  // check for memory shortage
+    // check for memory shortage
 
-  if (!mem_alloc_success)
-    {
-      return DE265_ERROR_OUT_OF_MEMORY;
-    }
+    if (!mem_alloc_success)
+      {
+        return DE265_ERROR_OUT_OF_MEMORY;
+      }
+  }
 
   //alloc_functions = *allocfunc;
   //alloc_userdata  = userdata;
@@ -278,6 +390,9 @@ de265_error de265_image::alloc_image(int w,int h, enum de265_chroma c,
     mem_alloc_success &= intraPredMode.alloc(sps->PicWidthInMinPUs, sps->PicHeightInMinPUs,
                                              sps->Log2MinPUSize);
 
+    mem_alloc_success &= intraPredModeC.alloc(sps->PicWidthInMinPUs, sps->PicHeightInMinPUs,
+                                              sps->Log2MinPUSize);
+
     // cb info
 
     mem_alloc_success &= cb_info.alloc(sps->PicWidthInMinCbsY, sps->PicHeightInMinCbsY,
@@ -349,9 +464,17 @@ void de265_image::release()
 
   if (pixels[0])
     {
-      image_allocation_functions.release_buffer(decctx, this,
-                                                decctx ? decctx->param_image_allocation_userdata : NULL);
-      
+      if (encoder_image_release_func != NULL) {
+        encoder_image_release_func(encctx, this,
+                                   encctx->param_image_allocation_userdata);
+      }
+      else {
+        image_allocation_functions.release_buffer(decctx, this,
+                                                decctx ?
+                                                  decctx->param_image_allocation_userdata :
+                                                  NULL);
+      }
+
       for (int i=0;i<3;i++)
         {
           pixels[i] = NULL;
@@ -393,7 +516,7 @@ de265_error de265_image::copy_image(const de265_image* src)
   */
 
   de265_error err = alloc_image(src->width, src->height, src->chroma_format, &src->sps, false,
-                                src->decctx, src->pts, src->user_data, false);
+                                src->decctx, src->encctx, src->pts, src->user_data, false);
   if (err != DE265_OK) {
     return err;
   }
@@ -412,33 +535,42 @@ void de265_image::copy_lines_from(const de265_image* src, int first, int end)
   assert(first % 2 == 0);
   assert(end   % 2 == 0);
 
+  int luma_bpp   = (sps.BitDepth_Y+7)/8;
+  int chroma_bpp = (sps.BitDepth_C+7)/8;
+
   if (src->stride == stride) {
-    memcpy(pixels[0]      + first*stride,
-           src->pixels[0] + first*src->stride,
-           (end-first)*stride);
+    memcpy(pixels[0]      + first*stride * luma_bpp,
+           src->pixels[0] + first*src->stride * luma_bpp,
+           (end-first)*stride * luma_bpp);
   }
   else {
     for (int yp=first;yp<end;yp++) {
-      memcpy(pixels[0]+yp*stride, src->pixels[0]+yp*src->stride, src->width);
+      memcpy(pixels[0]+yp*stride * luma_bpp,
+             src->pixels[0]+yp*src->stride * luma_bpp,
+             src->width * luma_bpp);
     }
   }
 
-  int first_chroma = first>>1;
-  int end_chroma   = end>>1;
+  int first_chroma = first / src->sps.SubHeightC;
+  int end_chroma   = end / src->sps.SubHeightC;
 
   if (src->chroma_format != de265_chroma_mono) {
     if (src->chroma_stride == chroma_stride) {
-      memcpy(pixels[1]      + first_chroma*chroma_stride,
-             src->pixels[1] + first_chroma*chroma_stride,
-             (end_chroma-first_chroma) * chroma_stride);
-      memcpy(pixels[2]      + first_chroma*chroma_stride,
-             src->pixels[2] + first_chroma*chroma_stride,
-             (end_chroma-first_chroma) * chroma_stride);
+      memcpy(pixels[1]      + first_chroma*chroma_stride * chroma_bpp,
+             src->pixels[1] + first_chroma*chroma_stride * chroma_bpp,
+             (end_chroma-first_chroma) * chroma_stride * chroma_bpp);
+      memcpy(pixels[2]      + first_chroma*chroma_stride * chroma_bpp,
+             src->pixels[2] + first_chroma*chroma_stride * chroma_bpp,
+             (end_chroma-first_chroma) * chroma_stride * chroma_bpp);
     }
     else {
       for (int y=first_chroma;y<end_chroma;y++) {
-        memcpy(pixels[1]+y*chroma_stride, src->pixels[1]+y*src->chroma_stride, src->chroma_width);
-        memcpy(pixels[2]+y*chroma_stride, src->pixels[2]+y*src->chroma_stride, src->chroma_width);
+        memcpy(pixels[1]+y*chroma_stride * chroma_bpp,
+               src->pixels[1]+y*src->chroma_stride * chroma_bpp,
+               src->chroma_width * chroma_bpp);
+        memcpy(pixels[2]+y*chroma_stride * chroma_bpp,
+               src->pixels[2]+y*src->chroma_stride * chroma_bpp,
+               src->chroma_width * chroma_bpp);
       }
     }
   }
@@ -463,14 +595,20 @@ void de265_image::thread_start(int nThreads)
 {
   de265_mutex_lock(&mutex);
 
+  //printf("nThreads before: %d %d\n",nThreadsQueued, nThreadsTotal);
+
   nThreadsQueued += nThreads;
   nThreadsTotal += nThreads;
 
+  //printf("nThreads after: %d %d\n",nThreadsQueued, nThreadsTotal);
+
   de265_mutex_unlock(&mutex);
 }
 
-void de265_image::thread_run()
+void de265_image::thread_run(const thread_task* task)
 {
+  //printf("run thread %s\n", task->name().c_str());
+
   de265_mutex_lock(&mutex);
   nThreadsQueued--;
   nThreadsRunning++;
@@ -493,8 +631,10 @@ void de265_image::thread_unblocks()
   de265_mutex_unlock(&mutex);
 }
 
-void de265_image::thread_finishes()
+void de265_image::thread_finishes(const thread_task* task)
 {
+  //printf("finish thread %s\n", task->name().c_str());
+
   de265_mutex_lock(&mutex);
 
   nThreadsRunning--;
@@ -517,6 +657,8 @@ void de265_image::wait_for_progress(thread_task* task, int ctbx,int ctby, int pr
 
 void de265_image::wait_for_progress(thread_task* task, int ctbAddrRS, int progress)
 {
+  if (task==NULL) { return; }
+
   de265_progress_lock* progresslock = &ctb_progress[ctbAddrRS];
   if (progresslock->get_progress() < progress) {
     thread_blocks();
@@ -558,7 +700,7 @@ void de265_image::clear_metadata()
   // during decoding (especially log2CbSize), but it is unlikely to be faster than the memset.
 
   cb_info.clear();
-  tu_info.clear();
+  //tu_info.clear();  // done on the fly
   ctb_info.clear();
   deblk_info.clear();
 
@@ -570,7 +712,7 @@ void de265_image::clear_metadata()
 }
 
 
-void de265_image::set_mv_info(int x,int y, int nPbW,int nPbH, const PredVectorInfo* mv)
+void de265_image::set_mv_info(int x,int y, int nPbW,int nPbH, const MotionVectorSpec& mv)
 {
   int log2PuSize = 2;
 
@@ -584,7 +726,7 @@ void de265_image::set_mv_info(int x,int y, int nPbW,int nPbH, const PredVectorIn
   for (int pby=0;pby<hPu;pby++)
     for (int pbx=0;pbx<wPu;pbx++)
       {
-        pb_info[ xPu+pbx + (yPu+pby)*stride ].mvi = *mv;
+        pb_info[ xPu+pbx + (yPu+pby)*stride ].mv = mv;
       }
 }
 
@@ -635,9 +777,9 @@ bool de265_image::available_pred_blk(int xC,int yC, int nCbS, int xP, int yP,
     availableN = available_zscan(xP,yP,xN,yN);
   }
   else {
-    availableN = !(nPbW<<1 == nCbS && nPbH<<1 == nCbS &&
+    availableN = !(nPbW<<1 == nCbS && nPbH<<1 == nCbS &&  // NxN
                    partIdx==1 &&
-                   yN >= yC+nPbH && xN < xC+nPbW);
+                   yN >= yC+nPbH && xN < xC+nPbW);  // xN/yN inside partIdx 2
   }
 
   if (availableN && get_pred_mode(xN,yN) == MODE_INTRA) {
@@ -646,4 +788,3 @@ bool de265_image::available_pred_blk(int xC,int yC, int nCbS, int xP, int yP,
 
   return availableN;
 }
-
