@@ -33,6 +33,7 @@
 #include <stdbool.h>
 #endif
 #include "libde265/de265.h"
+#include "libde265/en265.h"
 #include "libde265/sps.h"
 #include "libde265/pps.h"
 #include "libde265/motion.h"
@@ -40,7 +41,6 @@
 #include "libde265/slice.h"
 #include "libde265/nal.h"
 
-
 enum PictureState {
   UnusedForReference,
   UsedForShortTermReference,
@@ -78,23 +78,30 @@ enum PictureState {
 #define CTB_PROGRESS_DEBLK_H   3
 #define CTB_PROGRESS_SAO       4
 
+class decoder_context;
+
 template <class DataUnit> class MetaDataArray
 {
  public:
   MetaDataArray() { data=NULL; data_size=0; log2unitSize=0; width_in_units=0; height_in_units=0; }
   ~MetaDataArray() { free(data); }
 
-  bool alloc(int w,int h, int _log2unitSize) {
+  LIBDE265_CHECK_RESULT bool alloc(int w,int h, int _log2unitSize) {
     int size = w*h;
 
     if (size != data_size) {
       free(data);
       data = (DataUnit*)malloc(size * sizeof(DataUnit));
+      if (data == NULL) {
+        data_size = 0;
+        return false;
+      }
       data_size = size;
-      width_in_units = w;
-      height_in_units = h;
     }
 
+    width_in_units = w;
+    height_in_units = h;
+
     log2unitSize = _log2unitSize;
 
     return data != NULL;
@@ -108,6 +115,9 @@ template <class DataUnit> class MetaDataArray
     int unitX = x>>log2unitSize;
     int unitY = y>>log2unitSize;
 
+    assert(unitX >= 0 && unitX < width_in_units);
+    assert(unitY >= 0 && unitY < height_in_units);
+
     return data[ unitX + unitY*width_in_units ];
   }
 
@@ -115,6 +125,9 @@ template <class DataUnit> class MetaDataArray
     int unitX = x>>log2unitSize;
     int unitY = y>>log2unitSize;
 
+    assert(unitX >= 0 && unitX < width_in_units);
+    assert(unitY >= 0 && unitY < height_in_units);
+
     return data[ unitX + unitY*width_in_units ];
   }
 
@@ -122,12 +135,17 @@ template <class DataUnit> class MetaDataArray
     int unitX = x>>log2unitSize;
     int unitY = y>>log2unitSize;
 
+    assert(unitX >= 0 && unitX < width_in_units);
+    assert(unitY >= 0 && unitY < height_in_units);
+
     data[ unitX + unitY*width_in_units ] = d;
   }
 
   DataUnit& operator[](int idx) { return data[idx]; }
   const DataUnit& operator[](int idx) const { return data[idx]; }
 
+  int size() const { return data_size; }
+
   // private:
   DataUnit* data;
   int data_size;
@@ -146,6 +164,16 @@ template <class DataUnit> class MetaDataArray
         cb_info[ cbx + cby*cb_info.width_in_units ].Field = value;  \
       }
 
+#define CLEAR_TB_BLK(x,y,log2BlkWidth)              \
+  int tuX = x >> tu_info.log2unitSize; \
+  int tuY = y >> tu_info.log2unitSize; \
+  int width = 1 << (log2BlkWidth - tu_info.log2unitSize);           \
+  for (int tuy=tuY;tuy<tuY+width;tuy++)                             \
+    for (int tux=tuX;tux<tuX+width;tux++)                           \
+      {                                                             \
+        tu_info[ tux + tuy*tu_info.width_in_units ] = 0;  \
+      }
+
 
 typedef struct {
   uint16_t SliceAddrRS;
@@ -153,31 +181,45 @@ typedef struct {
 
   sao_info saoInfo;
   bool     deblock;         // this CTB has to be deblocked
-  bool     has_pcm;         // pcm is used in this CTB
-  bool     has_cu_transquant_bypass; // transquant_bypass is used in this CTB
+
+  // The following flag helps to quickly check whether we have to
+  // check all conditions in the SAO filter or whether we can skip them.
+  bool     has_pcm_or_cu_transquant_bypass; // pcm or transquant_bypass is used in this CTB
 } CTB_info;
 
 
 typedef struct {
-  uint8_t log2CbSize : 3;   // [0;6] (1<<log2CbSize) = 64
+  uint8_t log2CbSize : 3;   /* [0;6] (1<<log2CbSize) = 64
+                               This is only set in the top-left corner of the CB.
+                               The other values should be zero.
+                               TODO: in the encoder, we have to clear to zero.
+                               Used in deblocking and QP-scale decoding */
   uint8_t PartMode : 3;     // (enum PartMode)  [0;7] set only in top-left of CB
-                            // TODO: could be removed if prediction-block-boundaries would be
-                            // set during decoding
-  uint8_t ctDepth : 2;      // [0:3]? (0:64, 1:32, 2:16, 3:8)
+                            // Used for spatial merging candidates in current frame
+                            // and for deriving interSplitFlag in decoding.
+
+  uint8_t ctDepth : 2;      // [0:3]? (for CTB size 64: 0:64, 1:32, 2:16, 3:8)
+                            // Used for decoding/encoding split_cu flag.
+
+  // --- byte boundary ---
   uint8_t PredMode : 2;     // (enum PredMode)  [0;2] must be saved for past images
-  uint8_t pcm_flag : 1;     //
-  uint8_t cu_transquant_bypass : 1;
+                            // Used in motion decoding.
+  uint8_t pcm_flag : 1;     // Stored for intra-prediction / SAO
+  uint8_t cu_transquant_bypass : 1; // Stored for SAO
+  // note: 4 bits left
 
-  int8_t  QP_Y;
+  // --- byte boundary ---
+  int8_t  QP_Y;  // Stored for QP prediction
 
-  // uint8_t pcm_flag;  // TODO
 } CB_ref_info;
 
 
 typedef struct {
-  PredVectorInfo mvi; // TODO: this can be done in 16x16 grid
+  MotionVectorSpec mv; // TODO: this can be done in 16x16 grid
 } PB_ref_info;
 
+// intraPredMode:   Used for determining scanIdx when decoding/encoding coefficients.
+
 
 
 struct de265_image {
@@ -185,9 +227,15 @@ struct de265_image {
   ~de265_image();
 
 
-  de265_error alloc_image(int w,int h, enum de265_chroma c, const seq_parameter_set* sps,
-                          bool allocMetadata, decoder_context* ctx, de265_PTS pts, void* user_data,
-                          bool isOutputImage);
+  de265_error alloc_image(int w,int h, enum de265_chroma c,
+                          const seq_parameter_set* sps,
+                          bool allocMetadata,
+                          decoder_context* dctx,
+                          class encoder_context* ectx,
+                          de265_PTS pts, void* user_data,
+                          bool useCustomAllocFunctions);
+
+  //de265_error alloc_encoder_data(const seq_parameter_set* sps);
 
   bool is_allocated() const { return pixels[0] != NULL; }
 
@@ -212,12 +260,35 @@ struct de265_image {
     return pixels[cIdx] + xpos + ypos*stride;
   }
 
+
+  /// xpos;ypos in actual plane resolution
+  template <class pixel_t>
+  pixel_t* get_image_plane_at_pos_NEW(int cIdx, int xpos,int ypos)
+  {
+    int stride = get_image_stride(cIdx);
+    return (pixel_t*)(pixels[cIdx] + (xpos + ypos*stride)*sizeof(pixel_t));
+  }
+
   const uint8_t* get_image_plane_at_pos(int cIdx, int xpos,int ypos) const
   {
     int stride = get_image_stride(cIdx);
     return pixels[cIdx] + xpos + ypos*stride;
   }
 
+  void* get_image_plane_at_pos_any_depth(int cIdx, int xpos,int ypos)
+  {
+    int stride = get_image_stride(cIdx);
+    return pixels[cIdx] + ((xpos + ypos*stride) << bpp_shift[cIdx]);
+  }
+
+  const void* get_image_plane_at_pos_any_depth(int cIdx, int xpos,int ypos) const
+  {
+    int stride = get_image_stride(cIdx);
+    return pixels[cIdx] + ((xpos + ypos*stride) << bpp_shift[cIdx]);
+  }
+
+  /* Number of pixels in one row (not number of bytes).
+   */
   int get_image_stride(int cIdx) const
   {
     if (cIdx==0) return stride;
@@ -232,6 +303,18 @@ struct de265_image {
 
   enum de265_chroma get_chroma_format() const { return chroma_format; }
 
+  int get_bit_depth(int cIdx) const {
+    if (cIdx==0) return sps.BitDepth_Y;
+    else         return sps.BitDepth_C;
+  }
+
+  int get_bytes_per_pixel(int cIdx) const {
+    return (get_bit_depth(cIdx)+7)/8;
+  }
+
+  bool high_bit_depth(int cIdx) const {
+    return get_bit_depth(cIdx)>8;
+  }
 
   bool can_be_released() const { return PicOutputFlag==false && PicState==UnusedForReference; }
 
@@ -251,11 +334,17 @@ struct de265_image {
 
   static de265_image_allocation default_image_allocation;
 
+  void printBlk(const char* title, int x0,int y0,int blkSize,int cIdx) const {
+    ::printBlk(title, get_image_plane_at_pos(cIdx,x0,y0),
+               blkSize, get_image_stride(cIdx));
+  }
+
 private:
   uint32_t ID;
   static uint32_t s_next_image_ID;
 
   uint8_t* pixels[3];
+  uint8_t  bpp_shift[3];  // 0 for 8 bit, 1 for 16 bit
 
   enum de265_chroma chroma_format;
 
@@ -291,12 +380,16 @@ public:
   seq_parameter_set   sps;  // the SPS used for decoding this image
   pic_parameter_set   pps;  // the PPS used for decoding this image
   decoder_context*    decctx;
+  class encoder_context*    encctx;
+
+  int number_of_ctbs() const { return ctb_info.size(); }
 
 private:
   MetaDataArray<CTB_info>    ctb_info;
   MetaDataArray<CB_ref_info> cb_info;
   MetaDataArray<PB_ref_info> pb_info;
   MetaDataArray<uint8_t>     intraPredMode;
+  MetaDataArray<uint8_t>     intraPredModeC;
   MetaDataArray<uint8_t>     tu_info;
   MetaDataArray<uint8_t>     deblk_info;
 
@@ -307,6 +400,9 @@ public:
   void*     user_data;
   void*     plane_user_data[3];  // this is logically attached to the pixel data pointers
   de265_image_allocation image_allocation_functions; // the functions used for memory allocation
+  void (*encoder_image_release_func)(en265_encoder_context*,
+                                     de265_image*,
+                                     void* userdata);
 
   uint8_t integrity; /* Whether an error occured while the image was decoded.
                         When generated, this is initialized to INTEGRITY_CORRECT,
@@ -328,12 +424,13 @@ public:
 
 
   void thread_start(int nThreads);
-  void thread_run();
+  void thread_run(const thread_task*);
   void thread_blocks();
   void thread_unblocks();
-  void thread_finishes(); /* NOTE: you should not access any data in the thread_task after
-                             calling this, as this function may unlock other threads that
-                             will push this image to the output queue and free all decoder data. */
+  /* NOTE: you should not access any data in the thread_task after
+     calling this, as this function may unlock other threads that
+     will push this image to the output queue and free all decoder data. */
+  void thread_finishes(const thread_task*);
 
   void wait_for_progress(thread_task* task, int ctbx,int ctby, int progress);
   void wait_for_progress(thread_task* task, int ctbAddrRS, int progress);
@@ -384,10 +481,12 @@ public:
     return get_pred_mode(x,y)==MODE_SKIP;
   }
 
-  void set_pcm_flag(int x,int y, int log2BlkWidth)
+  void set_pcm_flag(int x,int y, int log2BlkWidth, uint8_t value=1)
   {
-    SET_CB_BLK(x,y,log2BlkWidth, pcm_flag, 1);
-    ctb_info.get(x,y).has_pcm = true;
+    SET_CB_BLK(x,y,log2BlkWidth, pcm_flag, value);
+
+    // TODO: in the encoder, we somewhere have to clear this
+    ctb_info.get(x,y).has_pcm_or_cu_transquant_bypass = true;
   }
 
   int  get_pcm_flag(int x,int y) const
@@ -395,10 +494,12 @@ public:
     return cb_info.get(x,y).pcm_flag;
   }
 
-  void set_cu_transquant_bypass(int x,int y, int log2BlkWidth)
+  void set_cu_transquant_bypass(int x,int y, int log2BlkWidth, uint8_t value=1)
   {
-    SET_CB_BLK(x,y,log2BlkWidth, cu_transquant_bypass, 1);
-    ctb_info.get(x,y).has_cu_transquant_bypass = true;
+    SET_CB_BLK(x,y,log2BlkWidth, cu_transquant_bypass, value);
+
+    // TODO: in the encoder, we somewhere have to clear this
+    ctb_info.get(x,y).has_pcm_or_cu_transquant_bypass = true;
   }
 
   int  get_cu_transquant_bypass(int x,int y) const
@@ -406,11 +507,16 @@ public:
     return cb_info.get(x,y).cu_transquant_bypass;
   }
 
-  void set_log2CbSize(int x0, int y0, int log2CbSize)
+  void set_log2CbSize(int x0, int y0, int log2CbSize, bool fill)
   {
-    cb_info.get(x0,y0).log2CbSize = log2CbSize;
+    // In theory, we could assume that remaining cb_info blocks are initialized to zero.
+    // But in corrupted streams, slices may overlap and set contradicting log2CbSizes.
+    // We also need this for encoding.
+    if (fill) {
+      SET_CB_BLK(x0,y0,log2CbSize, log2CbSize, 0);
+    }
 
-    // assume that remaining cb_info blocks are initialized to zero
+    cb_info.get(x0,y0).log2CbSize = log2CbSize;
   }
 
   int  get_log2CbSize(int x0, int y0) const
@@ -461,6 +567,11 @@ public:
     tu_info.get(x0,y0) |= (1<<trafoDepth);
   }
 
+  void clear_split_transform_flags(int x0,int y0,int log2CbSize)
+  {
+    CLEAR_TB_BLK (x0,y0, log2CbSize);
+  }
+
   int  get_split_transform_flag(int x0,int y0,int trafoDepth) const
   {
     return (tu_info.get(x0,y0) & (1<<trafoDepth));
@@ -500,12 +611,74 @@ public:
   void set_IntraPredMode(int PUidx,int log2blkSize, enum IntraPredMode mode)
   {
     int pbSize = 1<<(log2blkSize - intraPredMode.log2unitSize);
-    
+
     for (int y=0;y<pbSize;y++)
       for (int x=0;x<pbSize;x++)
         intraPredMode[PUidx + x + y*intraPredMode.width_in_units] = mode;
   }
 
+  void set_IntraPredMode(int x0,int y0,int log2blkSize,
+                         enum IntraPredMode mode)
+  {
+    int pbSize = 1<<(log2blkSize - intraPredMode.log2unitSize);
+    int PUidx  = (x0>>sps.Log2MinPUSize) + (y0>>sps.Log2MinPUSize)*sps.PicWidthInMinPUs;
+
+    for (int y=0;y<pbSize;y++)
+      for (int x=0;x<pbSize;x++) {
+        assert(x<sps.PicWidthInMinPUs);
+        assert(y<sps.PicHeightInMinPUs);
+
+        int idx = PUidx + x + y*intraPredMode.width_in_units;
+        assert(idx<intraPredMode.data_size);
+        intraPredMode[idx] = mode;
+      }
+  }
+
+
+  enum IntraPredMode get_IntraPredModeC(int x,int y) const
+  {
+    return (enum IntraPredMode)(intraPredModeC.get(x,y) & 0x3f);
+  }
+
+  bool is_IntraPredModeC_Mode4(int x,int y) const
+  {
+    return intraPredModeC.get(x,y) & 0x80;
+  }
+
+  void set_IntraPredModeC(int x0,int y0,int log2blkSize, enum IntraPredMode mode,
+                          bool is_mode4)
+  {
+    uint8_t combinedValue = mode;
+    if (is_mode4) combinedValue |= 0x80;
+
+    int pbSize = 1<<(log2blkSize - intraPredMode.log2unitSize);
+    int PUidx  = (x0>>sps.Log2MinPUSize) + (y0>>sps.Log2MinPUSize)*sps.PicWidthInMinPUs;
+
+    for (int y=0;y<pbSize;y++)
+      for (int x=0;x<pbSize;x++) {
+        assert(x<sps.PicWidthInMinPUs);
+        assert(y<sps.PicHeightInMinPUs);
+
+        int idx = PUidx + x + y*intraPredModeC.width_in_units;
+        assert(idx<intraPredModeC.data_size);
+        intraPredModeC[idx] = combinedValue;
+      }
+  }
+
+
+  /*
+  // NOTE: encoder only
+  void set_ChromaIntraPredMode(int x,int y,int log2BlkWidth, enum IntraChromaPredMode mode)
+  {
+    SET_CB_BLK (x, y, log2BlkWidth, intra_chroma_pred_mode, mode);
+  }
+
+  // NOTE: encoder only
+  enum IntraChromaPredMode get_ChromaIntraPredMode(int x,int y) const
+  {
+    return (enum IntraChromaPredMode)(cb_info.get(x,y).intra_chroma_pred_mode);
+  }
+  */
 
   // --- CTB metadata access ---
 
@@ -547,30 +720,42 @@ public:
     return ctb_info[ctb].SliceHeaderIndex;
   }
 
+  bool is_SliceHeader_available(int x,int y) const
+  {
+    int idx = ctb_info.get(x,y).SliceHeaderIndex;
+    return idx >= 0 && idx < slices.size();
+  }
+
   slice_segment_header* get_SliceHeader(int x, int y)
   {
-    return slices[ get_SliceHeaderIndex(x,y) ];
+    int idx = get_SliceHeaderIndex(x,y);
+    if (idx >= slices.size()) { return NULL; }
+    return slices[idx];
   }
 
   slice_segment_header* get_SliceHeaderCtb(int ctbX, int ctbY)
   {
-    return slices[ get_SliceHeaderIndexCtb(ctbX,ctbY) ];
+    int idx = get_SliceHeaderIndexCtb(ctbX,ctbY);
+    if (idx >= slices.size()) { return NULL; }
+    return slices[idx];
   }
 
   const slice_segment_header* get_SliceHeaderCtb(int ctbX, int ctbY) const
   {
-    return slices[ get_SliceHeaderIndexCtb(ctbX,ctbY) ];
+    int idx = get_SliceHeaderIndexCtb(ctbX,ctbY);
+    if (idx >= slices.size()) { return NULL; }
+    return slices[idx];
   }
-  
+
   void set_sao_info(int ctbX,int ctbY,const sao_info* saoinfo)
   {
     sao_info* sao = &ctb_info[ctbX + ctbY*ctb_info.width_in_units].saoInfo;
-    
+
     memcpy(sao,
            saoinfo,
            sizeof(sao_info));
   }
-  
+
   const sao_info* get_sao_info(int ctbX,int ctbY) const
   {
     return &ctb_info[ctbX + ctbY*ctb_info.width_in_units].saoInfo;
@@ -589,16 +774,10 @@ public:
   }
 
 
-  bool get_CTB_has_pcm(int ctbX,int ctbY) const
+  bool get_CTB_has_pcm_or_cu_transquant_bypass(int ctbX,int ctbY) const
   {
     int idx = ctbX + ctbY*ctb_info.width_in_units;
-    return ctb_info[idx].has_pcm;
-  }
-
-  bool get_CTB_has_cu_transquant_bypass(int ctbX,int ctbY) const
-  {
-    int idx = ctbX + ctbY*ctb_info.width_in_units;
-    return ctb_info[idx].has_cu_transquant_bypass;
+    return ctb_info[idx].has_pcm_or_cu_transquant_bypass;
   }
 
 
@@ -612,7 +791,7 @@ public:
   {
     const int xd = x0/4;
     const int yd = y0/4;
-    
+
     if (xd<deblk_info.width_in_units &&
         yd<deblk_info.height_in_units) {
       deblk_info[xd + yd*deblk_info.width_in_units] |= flags;
@@ -642,15 +821,16 @@ public:
 
   // --- PB metadata access ---
 
-  const PredVectorInfo* get_mv_info(int x,int y) const
+  const MotionVectorSpec* get_mv_info(int x,int y) const
   {
-    return &pb_info.get(x,y).mvi;
+    return &pb_info.get(x,y).mv;
   }
 
-  void set_mv_info(int x,int y, int nPbW,int nPbH, const PredVectorInfo* mv);
+  void set_mv_info(int x,int y, int nPbW,int nPbH, const MotionVectorSpec& mv);
 
-// --- value logging ---
+  // --- value logging ---
 
+  void printBlk(int x0,int y0, int cIdx, int log2BlkSize);
 };
 
 
