@@ -31,17 +31,37 @@
 #endif
 
 
-static void compute_NumPoc(ref_pic_set* rpset)
+void ref_pic_set::reset()
 {
-  rpset->NumPocTotalCurr_shortterm_only = 0;
-  
-  for (int i=0; i<rpset->NumNegativePics; i++)
-    if (rpset->UsedByCurrPicS0[i])
-      rpset->NumPocTotalCurr_shortterm_only++;
+  NumNegativePics = 0;
+  NumPositivePics = 0;
+  NumDeltaPocs = 0;
+  NumPocTotalCurr_shortterm_only = 0;
+
+  for (int i=0;i<MAX_NUM_REF_PICS;i++) {
+    DeltaPocS0[i] = 0;
+    DeltaPocS1[i] = 0;
+
+    UsedByCurrPicS0[i] = 0;
+    UsedByCurrPicS1[i] = 0;
+  }
+}
+
+
+void ref_pic_set::compute_derived_values()
+{
+  NumPocTotalCurr_shortterm_only = 0;
+
+  for (int i=0; i<NumNegativePics; i++)
+    if (UsedByCurrPicS0[i])
+      NumPocTotalCurr_shortterm_only++;
+
+  for (int i=0; i<NumPositivePics; i++)
+    if (UsedByCurrPicS1[i])
+      NumPocTotalCurr_shortterm_only++;
+
+  NumDeltaPocs = NumNegativePics + NumPositivePics;
 
-  for (int i=0; i<rpset->NumPositivePics; i++)
-    if (rpset->UsedByCurrPicS1[i])
-      rpset->NumPocTotalCurr_shortterm_only++;
 
   /*
     NOTE: this is done when reading the slice header.
@@ -62,7 +82,7 @@ static void compute_NumPoc(ref_pic_set* rpset)
    When coding the ref-pic-sets in the SPS, predicition is always from the previous set.
    In the slice header, the ref-pic-set can use any previous set as reference.
  */
-bool read_short_term_ref_pic_set(decoder_context* ctx,
+bool read_short_term_ref_pic_set(error_queue* errqueue,
                                  const seq_parameter_set* sps,
                                  bitreader* br,
                                  ref_pic_set* out_set, // where to store the read set
@@ -92,6 +112,14 @@ bool read_short_term_ref_pic_set(decoder_context* ctx,
     int delta_idx;
     if (sliceRefPicSet) { // idxRps == num_short_term_ref_pic_sets) {
       delta_idx = vlc = get_uvlc(br);
+      if (delta_idx==UVLC_ERROR) {
+        return false;
+      }
+
+      if (delta_idx>=idxRps) {
+        return false;
+      }
+
       delta_idx++;
     } else {
       delta_idx = 1;
@@ -102,6 +130,7 @@ bool read_short_term_ref_pic_set(decoder_context* ctx,
 
     int delta_rps_sign = get_bits(br,1);
     int abs_delta_rps  = vlc = get_uvlc(br);
+    if (vlc==UVLC_ERROR) { return false; }
     abs_delta_rps++;
     int DeltaRPS = (delta_rps_sign ? -abs_delta_rps : abs_delta_rps);
 
@@ -142,8 +171,13 @@ bool read_short_term_ref_pic_set(decoder_context* ctx,
 
     // positive list
     for (int j=nPositiveRIdx-1;j>=0;j--) {
+      assert(RIdx >= 0 && RIdx < sets.size());
+      assert(j>=0 && j < MAX_NUM_REF_PICS);
+
       int dPoc = sets[RIdx].DeltaPocS1[j] + DeltaRPS; // new delta
       if (dPoc<0 && use_delta_flag[nNegativeRIdx+j]) {
+        if (i>= MAX_NUM_REF_PICS) { return false; }
+
         out_set->DeltaPocS0[i] = dPoc;
         out_set->UsedByCurrPicS0[i] = used_by_curr_pic_flag[nNegativeRIdx+j];
         i++;
@@ -152,6 +186,8 @@ bool read_short_term_ref_pic_set(decoder_context* ctx,
 
     // frame 0
     if (DeltaRPS<0 && use_delta_flag[nDeltaPocsRIdx]) {
+      if (i>= MAX_NUM_REF_PICS) { return false; }
+
       out_set->DeltaPocS0[i] = DeltaRPS;
       out_set->UsedByCurrPicS0[i] = used_by_curr_pic_flag[nDeltaPocsRIdx];
       i++;
@@ -161,6 +197,8 @@ bool read_short_term_ref_pic_set(decoder_context* ctx,
     for (int j=0;j<nNegativeRIdx;j++) {
       int dPoc = sets[RIdx].DeltaPocS0[j] + DeltaRPS;
       if (dPoc<0 && use_delta_flag[j]) {
+        if (i>= MAX_NUM_REF_PICS) { return false; }
+
         out_set->DeltaPocS0[i] = dPoc;
         out_set->UsedByCurrPicS0[i] = used_by_curr_pic_flag[j];
         i++;
@@ -179,6 +217,8 @@ bool read_short_term_ref_pic_set(decoder_context* ctx,
     for (int j=nNegativeRIdx-1;j>=0;j--) {
       int dPoc = sets[RIdx].DeltaPocS0[j] + DeltaRPS;
       if (dPoc>0 && use_delta_flag[j]) {
+        if (i>= MAX_NUM_REF_PICS) { return false; }
+
         out_set->DeltaPocS1[i] = dPoc;
         out_set->UsedByCurrPicS1[i] = used_by_curr_pic_flag[j];
         i++;
@@ -187,6 +227,8 @@ bool read_short_term_ref_pic_set(decoder_context* ctx,
 
     // frame 0
     if (DeltaRPS>0 && use_delta_flag[nDeltaPocsRIdx]) {
+      if (i>= MAX_NUM_REF_PICS) { return false; }
+
       out_set->DeltaPocS1[i] = DeltaRPS;
       out_set->UsedByCurrPicS1[i] = used_by_curr_pic_flag[nDeltaPocsRIdx];
       i++;
@@ -196,6 +238,8 @@ bool read_short_term_ref_pic_set(decoder_context* ctx,
     for (int j=0;j<nPositiveRIdx;j++) {
       int dPoc = sets[RIdx].DeltaPocS1[j] + DeltaRPS;
       if (dPoc>0 && use_delta_flag[nNegativeRIdx+j]) {
+        if (i>= MAX_NUM_REF_PICS) { return false; }
+
         out_set->DeltaPocS1[i] = dPoc;
         out_set->UsedByCurrPicS1[i] = used_by_curr_pic_flag[nNegativeRIdx+j];
         i++;
@@ -204,8 +248,6 @@ bool read_short_term_ref_pic_set(decoder_context* ctx,
 
     out_set->NumPositivePics = i;
 
-    out_set->NumDeltaPocs = out_set->NumNegativePics + out_set->NumPositivePics;
-
   } else {
 
     // --- first, read the number of past and future frames in this set ---
@@ -222,14 +264,18 @@ bool read_short_term_ref_pic_set(decoder_context* ctx,
       out_set->NumDeltaPocs = 0;
       out_set->NumPocTotalCurr_shortterm_only = 0;
 
-      ctx->add_warning(DE265_WARNING_MAX_NUM_REF_PICS_EXCEEDED, false);
+      errqueue->add_warning(DE265_WARNING_MAX_NUM_REF_PICS_EXCEEDED, false);
       return false;
     }
 
+    if (num_negative_pics > MAX_NUM_REF_PICS ||
+        num_positive_pics > MAX_NUM_REF_PICS) {
+      errqueue->add_warning(DE265_WARNING_MAX_NUM_REF_PICS_EXCEEDED, false);
+      return false;
+    }
 
     out_set->NumNegativePics = num_negative_pics;
     out_set->NumPositivePics = num_positive_pics;
-    out_set->NumDeltaPocs = num_positive_pics + num_negative_pics;
 
     // --- now, read the deltas between the reference frames to fill the lists ---
 
@@ -237,7 +283,9 @@ bool read_short_term_ref_pic_set(decoder_context* ctx,
 
     int lastPocS=0;
     for (int i=0;i<num_negative_pics;i++) {
-      int  delta_poc_s0 = get_uvlc(br)+1;
+      int  delta_poc_s0 = get_uvlc(br);
+      if (delta_poc_s0==UVLC_ERROR) { return false; }
+      delta_poc_s0++;
       char used_by_curr_pic_s0_flag = get_bits(br,1);
 
       out_set->DeltaPocS0[i]      = lastPocS - delta_poc_s0;
@@ -249,7 +297,9 @@ bool read_short_term_ref_pic_set(decoder_context* ctx,
 
     lastPocS=0;
     for (int i=0;i<num_positive_pics;i++) {
-      int  delta_poc_s1 = get_uvlc(br)+1;
+      int  delta_poc_s1 = get_uvlc(br);
+      if (delta_poc_s1==UVLC_ERROR) { return false; }
+      delta_poc_s1++;
       char used_by_curr_pic_s1_flag = get_bits(br,1);
 
       out_set->DeltaPocS1[i]      = lastPocS + delta_poc_s1;
@@ -259,12 +309,76 @@ bool read_short_term_ref_pic_set(decoder_context* ctx,
   }
 
 
-  compute_NumPoc(out_set);
+  out_set->compute_derived_values();
 
   return true;
 }
 
 
+bool write_short_term_ref_pic_set_nopred(error_queue* errqueue,
+                                         const seq_parameter_set* sps,
+                                         CABAC_encoder& out,
+                                         const ref_pic_set* in_set, // which set to write
+                                         int idxRps,  // index of the set to be written
+                                         const std::vector<ref_pic_set>& sets, // previously read sets
+                                         bool sliceRefPicSet) // is this in the slice header?
+{
+  if (idxRps != 0) {
+    // inter_ref_pic_set_prediction_flag
+    out.write_bit(0);
+  }
+
+
+  // --- first, write the number of past and future frames in this set ---
+
+  out.write_uvlc(in_set->NumNegativePics);
+  out.write_uvlc(in_set->NumPositivePics);
+
+  // --- now, write the deltas between the reference frames to fill the lists ---
+
+  // past frames
+
+  int lastPocS=0;
+  for (int i=0;i<in_set->NumNegativePics;i++) {
+    int  delta_poc_s0 = lastPocS - in_set->DeltaPocS0[i];
+    char used_by_curr_pic_s0_flag = in_set->UsedByCurrPicS0[i];
+
+    assert(delta_poc_s0 >= 1);
+    out.write_uvlc(delta_poc_s0-1);
+    out.write_bit(used_by_curr_pic_s0_flag);
+    lastPocS = in_set->DeltaPocS0[i];
+  }
+
+  // future frames
+
+  lastPocS=0;
+  for (int i=0;i<in_set->NumPositivePics;i++) {
+    int  delta_poc_s1 = in_set->DeltaPocS1[i] - lastPocS;
+    char used_by_curr_pic_s1_flag = in_set->UsedByCurrPicS1[i];
+
+    assert(delta_poc_s1 >= 1);
+    out.write_uvlc(delta_poc_s1-1);
+    out.write_bit(used_by_curr_pic_s1_flag);
+    lastPocS = in_set->DeltaPocS1[i];
+  }
+
+  return true;
+}
+
+
+bool write_short_term_ref_pic_set(error_queue* errqueue,
+                                  const seq_parameter_set* sps,
+                                  CABAC_encoder& out,
+                                  const ref_pic_set* in_set, // which set to write
+                                  int idxRps,  // index of the set to be read
+                                  const std::vector<ref_pic_set>& sets, // previously read sets
+                                  bool sliceRefPicSet) // is this in the slice header?
+{
+  return write_short_term_ref_pic_set_nopred(errqueue, sps, out, in_set, idxRps, sets,
+                                             sliceRefPicSet);
+}
+
+
 void dump_short_term_ref_pic_set(const ref_pic_set* set, FILE* fh)
 {
   log2fh(fh,"NumDeltaPocs: %d [-:%d +:%d]\n", set->NumDeltaPocs,
