@@ -292,7 +292,7 @@ void draw_PB_block(const de265_image* srcimg,uint8_t* img,int stride,
     tint_rect(img,stride, x0,y0,w,h, cols[predMode], pixelSize);
   }
   else if (what == PBMotionVectors) {
-    const PredVectorInfo* mvi = srcimg->get_mv_info(x0,y0);
+    const MotionVectorSpec* mvi = srcimg->get_mv_info(x0,y0);
     int x = x0+w/2;
     int y = y0+h/2;
     if (mvi->predFlag[0]) {
@@ -554,4 +554,3 @@ LIBDE265_API void draw_Tiles(const de265_image* img, uint8_t* dst, int stride, i
     }
   }
 }
-
